import numpy as np
from numpy import typing as npt
from scipy import stats
from statsmodels.stats.multitest import multipletests
import pandas as pd
from typing import Tuple, Optional


def _wilcoxon_signed_rank_test(
    x: npt.NDArray,
    alternative: str,
):
    _, pvalue = stats.wilcoxon(x, alternative=alternative)
    return pvalue


def _t_test(
    x: npt.NDArray,
    alternative: str,
):
    _, pvalue = stats.ttest_1samp(x, popmean=0, alternative=alternative)
    return pvalue


def _sign_test(
    x: npt.NDArray,
    alternative: str,
):
    pos = np.sum(x >= 0)
    neg = np.sum(x <= 0)
    n = len(x)
    if alternative == "greater":
        pvalue = stats.binom_test(pos, n, p=0.5)
    elif alternative == "less":
        pvalue = stats.binom_test(neg, n, p=0.5)
    else:
        raise ValueError("alternative must be one of 'greater' or 'less'")
    return pvalue


def significance_tests(
    df: pd.DataFrame,
    alpha: float,
    test: str,
    alternative: str,
    multiple_testing: Optional[str] = None,
) -> Tuple[dict[str, float], dict[str, float]]:
    """
    Perform significance tests on a DataFrame of method results.
    The results are assumed to be differences to the baseline method.
    The function will perform a significance test for each method, comparing
    the differences to the baseline method to zero. The effect sizes will be
    calculated as the probability of superiority (i.e. the probability that a
    random observation from the method is greater than a random observation from
    the baseline method).

    Args:
        df (pd.DataFrame): DataFrame containing the method results.
        higher_is_better (bool): Flag indicating whether higher values are better.
        alpha (float): Significance level for the tests.
        test (str): Type of test to perform. Can be one of "wilcoxon", "paired_t_test", or "sign_test".
        alternative (str): The alternative hypothesis for the test. Can be one of "greater" or "less".
            Use "greater" if higher values are better, and "less" if lower values are better.
        multiple_testing (Optional[str], optional): Method for multiple testing correction.
            Can be one of "bonferroni", "fdr_bh", or None. Defaults to None.

    Returns:
        Tuple[dict[str, float], dict[str, float]]: A tuple containing two dictionaries.
            The first dictionary contains the effect sizes for each method.
            The second dictionary contains the p-values for each method.

    Raises:
        AssertionError: If `multiple_testing` or `test` is not one of the allowed values.
    """
    assert multiple_testing in ["bonferroni", "fdr_bh", None]
    assert test in ["wilcoxon", "t_test", "sign_test"]
    pvalues, effect_sizes = {}, {}
    for method_name in df:
        method_results = df[method_name].to_numpy()
        if test == "wilcoxon":
            pvalue = _wilcoxon_signed_rank_test(method_results, alternative)
        elif test == "t_test":
            pvalue = _t_test(method_results, alternative)
        elif test == "sign_test":
            pvalue = _sign_test(method_results, alternative)

        pvalues[method_name] = pvalue
        # Use probability of superiority as effect size
        effect_sizes[method_name] = np.mean(method_results > 0) if alternative == "greater" else np.mean(method_results < 0)

    if multiple_testing is not None:
        # Build list of method names and p-values in the same order
        # to make sure they are matched up correctly after correction using
        # statsmodels.stats.multitest.multipletests
        method_names = []
        pvalues_list = []
        for key in pvalues:
            method_names.append(key)
            pvalues_list.append(pvalues[key])

        _, pvals_corrected, _, _ = multipletests(
            pvalues_list,
            method=multiple_testing,
            alpha=alpha,
        )

        for i, method_name in enumerate(method_names):
            pvalues[method_name] = pvals_corrected[i]

    return effect_sizes, pvalues


def rowwise_pearsonr(
    a: npt.NDArray,
    b: npt.NDArray,
) -> Tuple[npt.NDArray, npt.NDArray]:
    """
    Calculates row-wise correlations between two arrays.
    This is a faster implementation of scipy.stats.pearsonr,
    as it only calculates correlation coefficients between corresponding rows,
    rather than between all pairs of rows.

    Parameters
    ----------
    a : npt.NDArray
        first set of row vectors (shape: [num_rows, num_measurements])
    b : npt.NDArray
        second set of row vectors (shape: [num_rows, num_measurements])

    Returns
    -------
    npt.NDArray
        row-wise Pearson correlations between a and b (shape: [num_rows])
    """
    # Subtract mean
    # [batch_size, num_observations]
    a -= a.mean(axis=1, keepdims=True)
    b -= b.mean(axis=1, keepdims=True)
    # Calculate numerator
    # [batch_size]
    cov = (a * b).sum(axis=1)
    # Calculate denominator
    # [batch_size]
    denom = np.sqrt((a**2).sum(axis=1)) * np.sqrt((b**2).sum(axis=1))
    # denom_zero = denom == 0.0
    # if np.any(denom_zero):
    #    warnings.warn(f"Zero standard deviation detected")

    # If the denominator is zero, that means one of the series is constant.
    # Correlation is technically undefined in this case, but covariance is 0.
    # We can just set the correlation coefficient to zero in this case.
    corrcoefs = np.divide(cov, denom, out=np.zeros_like(cov), where=denom != 0)

    return corrcoefs


def rowwise_spearmanr(
    a: npt.NDArray,
    b: npt.NDArray,
) -> Tuple[npt.NDArray, npt.NDArray]:
    """
    Calculates row-wise Spearman correlations between two arrays.
    This is a faster implementation of scipy.stats.spearmanr,
    as it only calculates correlation coefficients between corresponding rows,
    rather than between all pairs of rows.

    Parameters
    ----------
    a : npt.NDArray
        first set of row vectors (shape: [num_rows, num_measurements])
    b : npt.NDArray
        second set of row vectors (shape: [num_rows, num_measurements])

    Returns
    -------
    npt.NDArray
        row-wise Spearman correlations between a and b (shape: [num_rows])
    """
    # Calculate rank of each row
    # [batch_size, num_observations]
    a_ranks = stats.rankdata(a, axis=1)
    b_ranks = stats.rankdata(b, axis=1)

    # Spearman rank correlation is simply Pearson correlation of ranks
    return rowwise_pearsonr(a_ranks, b_ranks)



from typing import Dict, Callable
from torch.nn.functional import softmax
from torch import sigmoid
import torch


ACTIVATION_FNS: Dict[str, Callable[[torch.Tensor], torch.Tensor]] = {
    "linear": lambda x: x,
    "softmax": lambda x: softmax(x, dim=1),
    "sigmoid": lambda x: sigmoid(x)
}



from abc import abstractmethod
import torch
from torch import nn


class AttributionMethod:
    """
    Wrapper class for attribution methods.
    An attribution method takes 2 arguments (input and target) and produces
    attributions in the shape of the input.
    To create a compatible attribution method, override the :meth:`__call__` method
    and optionally the :meth:`__init__` method for kwargs.
    """
    def __init__(self, model: nn.Module, **kwargs) -> None:
        """
        Parameters
        ----------
        model : nn.Module
            Model to compute attributions for.
        """
        self.model = model

    @abstractmethod
    def __call__(self, batch_x: torch.Tensor,
                 batch_target: torch.Tensor) -> torch.Tensor:
        """Compute attributions for a batch of inputs.

        Parameters
        ----------
        batch_x : torch.Tensor
            Input samples.
        batch_target : torch.Tensor
            Targets to compute attributions for.
            Note that these need not be the same as the ground truth targets.

        Returns
        -------
        torch.Tensor
            Attributions for the given inputs and targets.

        """
        raise NotImplementedError



from torch import nn
from abc import abstractmethod
import copy


class ModelFactory:
    """
    Basic interface for a callable that returns a model.
    This is necessary for multi-GPU mode, as each subprocess
    needs its own copy of the model.
    """

    @abstractmethod
    def __call__(self) -> nn.Module:
        """Return a model.

        Returns
        -------
        nn.Module
            The model.
        """
        raise NotImplementedError


class BasicModelFactory(ModelFactory):
    """
    Basic implementation of a :class:`ModelFactory` that just returns a deep copy
    of a given model. This can be used for simple models, but requires
    the model to be picklable (e.g. lambda layers won't work and require
    a specific implementation of :class:`ModelFactory`)
    """

    def __init__(self, base_model: nn.Module) -> None:
        """
        Parameters
        ----------
        base_model : nn.Module
            Model of which to return a deep copy.
        """
        self.base_model = base_model

    def __call__(self) -> nn.Module:
        """Return a deep copy of the base model.

        Returns
        -------
        nn.Module
            Deep copy of the base model.
        """
        return copy.deepcopy(self.base_model)



import numpy as np
import torch
from skimage.segmentation import slic


def isin(a: torch.Tensor, b: torch.Tensor):
    # https://stackoverflow.com/questions/60918304/get-indices-of-elements-in-tensor-a-that-are-present-in-tensor-b
    return (a[..., None] == b).any(-1)


def segment_samples(samples: np.ndarray) -> np.ndarray:
    # Segment images using SLIC
    seg_images = np.stack(
        [
            slic(
                np.transpose(samples[i, ...], (1, 2, 0)),
                start_label=0,
                slic_zero=True,
            )
            for i in range(samples.shape[0])
        ]
    )
    seg_images = np.expand_dims(seg_images, axis=1)
    return seg_images


def segment_attributions(
    seg_images: np.ndarray, attrs: np.ndarray
) -> np.ndarray:
    segments = np.unique(seg_images)
    seg_img_flat = seg_images.reshape(seg_images.shape[0], -1)
    attrs_flat = attrs.reshape(attrs.shape[0], -1)
    result = np.zeros((seg_images.shape[0], len(segments)))
    for seg in segments:
        mask = (seg_img_flat == seg).astype(int)
        masked_attrs = mask * attrs_flat
        mask_size = np.sum(mask, axis=1)
        sum_attrs = np.sum(masked_attrs, axis=1)
        mean_attrs = np.divide(
            sum_attrs,
            mask_size,
            out=np.zeros_like(sum_attrs),
            where=mask_size != 0,
        )
        # If seg does not exist for image, mask_size == 0. Set to -inf.
        mean_attrs[mask_size == 0] = -np.inf
        result[:, seg] = mean_attrs
    return result



from typing import Dict, Type, Union, Tuple, List
from torch import nn

from attribench._attribution_method import AttributionMethod


ConfigDict = Dict[
        str,
        Union[
            Type[AttributionMethod],
            Tuple[Type[AttributionMethod], Dict],
        ],
    ]
"""
A ConfigDict is a dictionary mapping strings to either an AttributionMethod,
or a tuple consisting of an AttributionMethod constructor and a dictionary of
keyword arguments to pass to the constructor.
"""


class MethodFactory:
    """
    This class accepts a config dictionary for attribution methods in its
    constructor, and will return a dictionary of ready-to-use AttributionMethod
    objects when called with a model (nn.Module) as argument. This allows
    the attribution methods to be instantiated in subprocesses, which is
    necessary for computing attributions on multiple GPUs, as the methods need
    access to the specific copy of the model for their process.

    The config dictionary should map strings to either AttributionMethod
    constructors, or tuples consisting of an AttributionMethod constructor
    and a dictionary of keyword arguments to pass to the constructor.

    Example::

        {
            "method1": AttributionMethod1,
            "method2": (AttributionMethod2, {"kwarg1": 1, "kwarg2": 2}),
        }

    """

    def __init__(self, config_dict: ConfigDict) -> None:
        """
        Parameters
        ----------
        config_dict : ConfigDict
            Dictionary mapping strings to either AttributionMethod constructors,
            or tuples consisting of an AttributionMethod constructor and a
            dictionary of keyword arguments to pass to the constructor.
        """
        self.config_dict = config_dict

    def __call__(self, model: nn.Module) -> Dict[str, AttributionMethod]:
        """Create dictionary mapping method names to AttributionMethod objects.

        Parameters
        ----------
        model : nn.Module
            Model to compute attributions for.

        Returns
        -------
        Dict[str, AttributionMethod]
            Dictionary mapping method names to AttributionMethod objects.
        """
        result: Dict[str, AttributionMethod] = {}
        for method_name, entry in self.config_dict.items():
            if isinstance(entry, Tuple):
                # Entry consists of constructor and kwargs
                constructor, kwargs = entry
                result[method_name] = constructor(model, **kwargs)
            else:
                # Constructor has no kwargs
                result[method_name] = entry(model)
        return result

    def __len__(self) -> int:
        return len(self.config_dict)

    def get_method_names(self) -> List[str]:
        return list(self.config_dict.keys())
